#include <httpd.h>
#include <http_protocol.h>
#include <http_config.h>
#include <syslog.h>

/*
 * Copyright 2008 ed neville 
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); 
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at 
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License. 
 *
 * http://www.s5h.net/
 *
 * $Id: mod_cidr.c 95 2008-02-11 21:04:50Z ed $
 * $URL: svn+ssh://fw/repos/code/apache2_modules/cidr/mod_cidr.c $
 */

#include "error.h"
#include "cdb.h"

#define BUFFER 8192

module AP_MODULE_DECLARE_DATA cidr_module;

typedef struct cidr_cfg {
	int cmode;	/* environment, directory, server, combo */
#define CONFIG_SERVER	1
#define CONFIG_DIRECTORY	2
	int config_rec_type;
#define CONFIG_CDB_FILE	1
	char *config_rec_data; /* pointer to the data, for the moment this should only be a string for the .cdb location */
} cidr_cfg;

static cidr_cfg *our_dir_config( const request_rec *r )
{
	return( ( cidr_cfg* )ap_get_module_config( r->per_dir_config, &cidr_module ) );
}


static void *cidr_dir_conf( apr_pool_t *pool, char *x ) {
	/**
	 * allocate some memory to be returned to caller as location of directory configuration 
	 * */
	void *dir = (void*)apr_pcalloc( pool, sizeof( cidr_cfg ) );
	return( (void*)dir );
}

static const char *cidr_cmd_func( cmd_parms *cmd, void *cfg,
	const char *data ) {
	syslog( LOG_DEBUG, __FUNCTION__ );
	return( NULL );
}

static const char *set_cidr_cdb_file( cmd_parms *cmd, void *cfg, const char *val ) {
	if( cfg == NULL ) {
		return( NULL );
	}
	((cidr_cfg*)cfg)->config_rec_data = val;
	return( NULL );
}

static const command_rec cidr_cmds[] =  {
	AP_INIT_TAKE1(
		"CIDR_CDB_File",
		set_cidr_cdb_file,
		NULL,
		OR_ALL,
		"CIDR CDB File location" ),
	{NULL}
};

/**
 * returns the number to the power of 2
 */
int cidr_power_of( int x, int y ) {
	int return_val = x, i;

	for( i=0 ; i<y ; i++ ) {
		return_val = return_val * x;
	}
	return( return_val );
}

/**
 * this function just takes a pointer to cdb memory and a file descriptor. it might be a good idea if in the future this does some form of error checking
 */

static int cidr_setup_cdb( struct cdb *c, int fd ) {
	cdb_init( c, fd );
	cdb_findstart( c );
	return(0);
}

static int cidr_hook( request_rec *r ) {
	int ipnum,i,k,datalen,datapos;
	unsigned int j = 0;
	char *ptr, mask[BUFFER],*w,*x,*y,*z;
	struct cdb c;
	char *temp = (char *)apr_palloc( r->pool, BUFFER );
	FILE *fp;
	char *file_location;
	cidr_cfg *cfg;

	cfg = our_dir_config( r );
	if( cfg == NULL ) {
		return( DECLINED );
	}

	file_location = cfg->config_rec_data;

	if( file_location == NULL ) {
		return( DECLINED );
	}

	fp = fopen( file_location, "r" );
	if( !fp ) {
		return( DECLINED );
	}

	if( cidr_setup_cdb( &c, fileno( fp ) ) ) {
		fclose(fp);
		return( DECLINED );
	}

	apr_snprintf( temp, BUFFER, "%s", r->connection->remote_ip );

	w = temp;

	x = strchr( w, '.' ); *x = '\0'; x++;
	y = strchr( x, '.' ); *y = '\0'; y++;
	z = strchr( y, '.' ); *z = '\0'; z++;

	ipnum = ( atoi(w) << 24 ) | ( atoi(x) << 16 ) | ( atoi(y) << 8 ) | ( atoi(z) );

	snprintf( temp, BUFFER, "Inbound: %u", ipnum );
	// syslog( LOG_DEBUG, temp );

	/**
	 * now lets try and find our cidr allocation
	 */

	for( k = 0 ; k < 32 ; k++ ) {
		int base = cidr_power_of( 2, k );
		unsigned int l = ipnum & ~( base - 1 );

		/**
		 * make the unsigned long a string for key lookup
		 */
		snprintf( temp, BUFFER, "%u", l );

		snprintf( mask, BUFFER, "%s", temp );

		i = cdb_find( &c, temp, strlen( temp ) );

		// syslog( LOG_DEBUG, temp );
		
		if( i == 0 ) {
			continue;
		}

		if( i == -1 ) {
			syslog( LOG_DEBUG, "Error reading cdb" );
			fclose(fp);
			return( DECLINED );
		}

		datapos = cdb_datapos( &c );
		datalen = cdb_datalen( &c );

		if( cdb_read( &c, temp, BUFFER, datapos ) == -1 ) {
			syslog( LOG_DEBUG, "Cannot read" );
			continue;
		}
	
		apr_table_setn( r->headers_in, "CIDR_RANGE", mask );

		ptr = temp;

		apr_table_setn( r->headers_in, "CIDR_MASK", ptr );
		
		ptr += strlen( ptr )+1;
		apr_table_setn( r->headers_in, "CIDR_RIR", ptr );

		ptr += strlen( ptr )+1;
		apr_table_setn( r->headers_in, "CIDR_CC", ptr );

		break;
	}
	
	cdb_free( &c );
	fclose(fp);

	return( OK );
}


static void cidr_module_hook( apr_pool_t *pool ) {
	ap_hook_header_parser( cidr_hook, NULL, NULL, APR_HOOK_MIDDLE );
}

module AP_MODULE_DECLARE_DATA cidr_module = {
	STANDARD20_MODULE_STUFF,
	cidr_dir_conf,
	NULL,
	NULL,
	NULL,
	cidr_cmds,
	cidr_module_hook
};


